"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
// @ts-ignore
const Browserslist = require("browserslist");
// @ts-ignore
const browserslist_useragent_1 = require("browserslist-useragent");
// @ts-ignore
const caniuse_lite_1 = require("caniuse-lite");
const i_caniuse_lite_1 = require("./i-caniuse-lite");
// tslint:disable:no-magic-numbers
/**
 * These browsers should be skipped when deciding which browsers to take into account
 * when generating a browserslist
 * @type {Set<string>}
 */
const SKIP_BROWSERS_PAYLOAD = [
    // caniuse.com has some issues with reporting android browsers
    "android",
    "and_chr",
    "and_qq",
    "and_uc"
];
const SKIP_BROWSERS = new Set(SKIP_BROWSERS_PAYLOAD);
/**
 * A Regular Expression that captures the part of a browser version that should be kept
 * @type {RegExp}
 */
const NORMALIZE_BROWSER_VERSION_REGEXP = /(?![\d.,]+-)-*(.*)/;
/**
 * A Map between features and browsers that has partial support for them but should be allowed anyway
 * @type {Map<string, string[]>}
 */
const PARTIAL_SUPPORT_ALLOWANCES = new Map([
    [
        "shadowdomv1",
        ["chrome", "safari", "ios_saf"]
    ],
    [
        "custom-elementsv1",
        ["chrome", "safari", "ios_saf"]
    ]
]);
/**
 * A function that will match a browserslist on the given user agent
 * @param {string} userAgent
 * @param {string[]} browserslist
 * @returns {boolean}
 */
function matchBrowserslistOnUserAgent(userAgent, browserslist) {
    return browserslist_useragent_1.matchesUA(userAgent, {
        browsers: browserslist,
        _allowHigherVersions: true,
        ignoreMinor: true,
        ignorePatch: true
    });
}
exports.matchBrowserslistOnUserAgent = matchBrowserslistOnUserAgent;
/**
 * Returns the input query, but extended with the given options
 * @param {string[]} query
 * @param {string|string[]} extendWith
 * @returns {string[]}
 */
function extendQueryWith(query, extendWith) {
    const normalizedExtendWith = Array.isArray(extendWith) ? extendWith : [extendWith];
    return [...new Set([
            ...query,
            ...normalizedExtendWith
        ])];
}
/**
 * Returns the input query, but extended with 'unreleased versions'
 * @param {string[]} query
 * @returns {string[]}
 */
function extendQueryWithUnreleasedVersions(query) {
    return extendQueryWith(query, "unreleased versions");
}
/**
 * Generates a Browserslist based on browser support for the given features
 * @param {string[]} features
 * @returns {string}
 */
function browsersWithSupportForFeatures(...features) {
    const baseQuery = browsersWithSupportForFeaturesCommon(">=", ...features);
    return extendQueryWithUnreleasedVersions(baseQuery);
}
exports.browsersWithSupportForFeatures = browsersWithSupportForFeatures;
/**
 * Returns true if the given browserslist support all of the given features
 * @param {string[]} browserslist
 * @param {string} features
 * @returns {boolean}
 */
function browserslistSupportsFeatures(browserslist, ...features) {
    // First, generate an ideal browserslist that would target the given features exactly
    const normalizedIdealBrowserslist = Browserslist(browsersWithSupportForFeatures(...features));
    // Now, normalize the input browserslist
    const normalizedInputBrowserslist = Browserslist(extendQueryWithUnreleasedVersions(browserslist));
    // Now, compare the two and see if they align. If they do, the input browserslist *does* support all of the given features.
    // They align if all members of the input browserslist are included in the ideal browserslist
    return normalizedInputBrowserslist.every(option => normalizedIdealBrowserslist.includes(option));
}
exports.browserslistSupportsFeatures = browserslistSupportsFeatures;
/**
 * Generates a Browserslist based on browsers that *doesn't* support the given features
 * @param {string[]} features
 * @returns {string}
 */
function browsersWithoutSupportForFeatures(...features) {
    return browsersWithSupportForFeaturesCommon("<", ...features);
}
exports.browsersWithoutSupportForFeatures = browsersWithoutSupportForFeatures;
/**
 * Normalizes the given ICaniuseLiteFeature
 * @param {ICaniuseLiteFeature} feature
 * @returns {ICaniuseLiteFeatureNormalized}
 */
function getCaniuseLiteFeatureNormalized(feature) {
    Object.keys(feature.stats).forEach((browser) => {
        const browserDict = feature.stats[browser];
        Object.entries(browserDict).forEach(([version, support]) => {
            const versionMatch = version.match(NORMALIZE_BROWSER_VERSION_REGEXP);
            const normalizedVersion = versionMatch == null ? version : versionMatch[1];
            let supportKind;
            if (support.startsWith("y")) {
                supportKind = i_caniuse_lite_1.CaniuseSupportKind.AVAILABLE;
            }
            else if (support.startsWith("n")) {
                supportKind = i_caniuse_lite_1.CaniuseSupportKind.UNAVAILABLE;
            }
            else if (support.startsWith("a")) {
                supportKind = i_caniuse_lite_1.CaniuseSupportKind.PARTIAL_SUPPORT;
            }
            else {
                supportKind = i_caniuse_lite_1.CaniuseSupportKind.PREFIXED;
            }
            // Delete the rewritten version
            if (version !== normalizedVersion) {
                delete browserDict[version];
            }
            browserDict[normalizedVersion] = supportKind;
        });
    });
    return feature;
}
/**
 * Gets the latest (stable) version of the given browser
 * @param {CaniuseLiteBrowser} browser
 * @returns {string}
 */
function getLatestVersionOfBrowser(browser) {
    const queryResultsMapped = new Map();
    // Generate the Browserslist query
    const queryResult = Browserslist([`>= 0%`, `unreleased versions`]);
    // First, organize the different versions of the browsers inside the Map
    queryResult.forEach(part => {
        const [currentBrowser, version] = part.split(" ");
        let versions = queryResultsMapped.get(currentBrowser);
        if (versions == null) {
            versions = [];
            queryResultsMapped.set(currentBrowser, versions);
        }
        const versionMatch = version.match(NORMALIZE_BROWSER_VERSION_REGEXP);
        const normalizedVersion = versionMatch == null ? version : versionMatch[1];
        versions.push(normalizedVersion);
    });
    return queryResultsMapped.get(browser)
        .sort(compareVersions)
        .slice(-1)[0];
}
/**
 * Gets the support from caniuse for the given feature
 * @param {string} feature
 * @returns {ICaniuseLiteFeatureNormalized}
 */
function getSupport(feature) {
    return getCaniuseLiteFeatureNormalized(caniuse_lite_1.feature(caniuse_lite_1.features[feature])).stats;
}
/**
 * Compares two versions, a and b
 * @param {string} a
 * @param {string} b
 * @returns {number}
 */
function compareVersions(a, b) {
    const normalizedA = isNaN(parseFloat(a)) ? a : parseFloat(a);
    const normalizedB = isNaN(parseFloat(b)) ? b : parseFloat(b);
    if (typeof normalizedA === "string" && typeof normalizedB !== "string") {
        return 1;
    }
    if (typeof normalizedB === "string" && typeof normalizedA !== "string") {
        return -1;
    }
    if (normalizedA < normalizedB)
        return -1;
    if (normalizedA > normalizedB)
        return 1;
    return 0;
}
/**
 * Gets the first version that matches the given CaniuseSupportKind
 * @param {CaniuseSupportKind} kind
 * @param {object} stats
 * @returns {string | undefined}
 */
function getFirstVersionWithSupportKind(kind, stats) {
    // Sort all keys of the object
    const sortedKeys = Object.keys(stats).sort(compareVersions);
    for (const key of sortedKeys) {
        if (stats[key] === kind) {
            return key;
        }
    }
    return undefined;
}
/**
 * Common logic for the functions that generate browserslists based on feature support
 * @param {ComparisonOperator} comparisonOperator
 * @param {string} features
 * @returns {string[]}
 */
function browsersWithSupportForFeaturesCommon(comparisonOperator, ...features) {
    // All of the generated browser maps
    const browserMaps = [];
    for (const feature of features) {
        const support = getSupport(feature);
        // A map between browser names and their required versions
        const browserMap = new Map();
        Object.entries(support)
            .filter(([browser]) => !SKIP_BROWSERS.has(browser))
            .forEach(([browser, stats]) => {
            const fullSupportVersion = getFirstVersionWithSupportKind(i_caniuse_lite_1.CaniuseSupportKind.AVAILABLE, stats);
            const partialSupportVersion = getFirstVersionWithSupportKind(i_caniuse_lite_1.CaniuseSupportKind.PARTIAL_SUPPORT, stats);
            if (fullSupportVersion != null) {
                browserMap.set(browser, fullSupportVersion);
            }
            // Otherwise, check if partial support exists and should be allowed
            if (partialSupportVersion != null) {
                // Get all partial support allowances for this specific feature
                const partialSupportMatch = PARTIAL_SUPPORT_ALLOWANCES.get(feature);
                if (partialSupportMatch != null) {
                    // Check if partial support exists for the browser
                    if (partialSupportMatch.includes(browser)) {
                        // If no full supported version exists or if the partial supported version has a lower version number than the full supported one, use that one instead
                        if (fullSupportVersion == null || compareVersions(partialSupportVersion, fullSupportVersion) < 0) {
                            browserMap.set(browser, partialSupportVersion);
                        }
                    }
                }
            }
            // Apply additional checks depending on the comparison operator
            switch (comparisonOperator) {
                case "<":
                case "<=":
                    // Add all browsers with no support whatsoever, or those that require prefixing or flags
                    browserMap.set(browser, getLatestVersionOfBrowser(browser));
            }
        });
        browserMaps.push(browserMap);
    }
    // Now, remove all browsers that isn't part of all generated browser maps
    for (const browserMap of browserMaps) {
        for (const browser of browserMap.keys()) {
            if (!browserMaps.every(map => map.has(browser))) {
                // Delete the browser if it isn't included in all of the browser maps
                browserMap.delete(browser);
            }
        }
    }
    // Now, prepare a combined browser map
    const combinedBrowserMap = new Map();
    for (const browserMap of browserMaps) {
        for (const [browser, version] of browserMap.entries()) {
            // Take the existing entry from the combined map
            const existingVersion = combinedBrowserMap.get(browser);
            // The browser should be set in the map if it has no entry already
            let shouldSet = existingVersion == null;
            // If it *has* an entry, also check if the version should be updated, depending on the comparison operator
            if (existingVersion != null) {
                const comparableVersion = isNaN(parseFloat(version)) ? version : parseFloat(version);
                const comparableExistingVersion = isNaN(parseFloat(existingVersion)) ? version : parseFloat(existingVersion);
                switch (comparisonOperator) {
                    case ">=":
                    case ">":
                        shouldSet = comparableVersion >= comparableExistingVersion;
                        break;
                    case "<=":
                    case "<":
                        shouldSet = comparableVersion <= comparableExistingVersion;
                        break;
                }
            }
            if (shouldSet) {
                // Set the version in the map
                combinedBrowserMap.set(browser, version);
            }
        }
    }
    // Finally, generate a string array of the browsers
    return Array.from(combinedBrowserMap.entries())
        .map(([browser, version]) => `${browser} ${comparisonOperator} ${version}`);
}
//# sourceMappingURL=browserslist-generator.js.map